{#
    sphinxdoc/layout.html
    ~~~~~~~~~~~~~~~~~~~~~

    Sphinx layout template for the sphinxdoc theme.

    :copyright: Copyright 2007-2014 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
#}
{%- extends "basic/layout.html" %}

{%- block extrahead %}
  <script type="text/x-mathjax-config">
     MathJax.Hub.Config({
        "TeX": {Macros: {AA : "{\\unicode{x212B}}"}},
        "HTML-CSS": {scale: 90}
  });</script>
{% endblock %}



{% block rootrellink %}
   <li>[<a href="{{ pathto('intro') }}">intro</a>|</li>
   <li><a href="{{ pathto('parameters') }}">parameters</a>|</li>
   <li><a href="{{ pathto('fitting') }}"> minimize</a>|</li>
   <li><a href="{{ pathto('model') }}"> model</a>|</li>
   <li><a href="{{ pathto('builtin_models') }}"> built-in models</a>|</li>
   <li><a href="{{ pathto('confidence') }}">confidence intervals</a>|</li>
   <li><a href="{{ pathto('bounds') }}">bounds</a>|</li>
   <li><a href="{{ pathto('constraints') }}">constraints</a>]</li>
{% endblock %}

{% block relbar1 %}
<div>
<table border=0>
  <tr><td></td><td width=85% padding=5 align=left>
       <a href="index.html" style="color: #157"> <font size=+3>LMFIT</font></a>
     </td>
     <td width=7% align=left>
         <a href="contents.html" style="color: #882222">
         <font size+=1>Contents</font></a> </td>
     <td width=7% align=left>
          <a href="installation.html" style="color: #882222">
          <font size+=1>Download</font></a></td>
     <td></td>
  </tr>
  <tr><td></td><td width=75% padding=5 align=left>
        <a href="index.html" style="color: #157"> <font size=+2>
	Non-Linear Least-Squares Minimization and Curve-Fitting for Python</font></a>
     </td>
     <td width=7% align=left>
         <a href="faq.html" style="color: #882222">
         <font size+=1>FAQ</font></a> </td>
     <td width=7% align=left>
        <a href="https://github.com/lmfit/lmfit-py/" style="color: #882222">
         <font size+=1>Develop</font></a></td>
     <td></td>
  </tr>
</table>
</div>
{{ super() }}
{% endblock %}

{# put the sidebar before the body #}
{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}
